# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from sys import exit
import os.path

import adles.vsphere.folder_utils as futils
from adles.utils import pad, read_json, is_folder, is_vm, get_vlan
from adles.vsphere import Vsphere
from adles.vsphere.network_utils import create_portgroup
from adles.vsphere.vm import VM
from adles.interfaces import Interface


class VsphereInterface(Interface):
    """ Generic interface for the VMware vSphere platform """
    __version__ = "1.0.3"

    # noinspection PyMissingConstructor
    def __init__(self, infra, spec):
        """
        NOTE: it is assumed that the infrastructure and spec are both valid,
        and thus checks on key existence and types are not performed for REQUIRED elements.
        :param dict infra: Infrastructure information
        :param dict spec: The parsed exercise specification
        """
        self._log = logging.getLogger('VsphereInterface')
        self._log.debug("Initializing VsphereInterface %s", VsphereInterface.__version__)

        self.spec = spec
        self.metadata = spec["metadata"]
        self.services = spec["services"]
        self.networks = spec["networks"]
        self.folders = spec["folders"]
        self.infra = infra
        self.master_folder = None
        self.template_folder = None
        self.net_table = {}  # Used to do lookups of Generic networks during deployment
        self.masters = {}  # Cache containing Master instances (TODO: potential naming conflicts)

        if "thresholds" in infra:
            self.thresholds = infra["thresholds"]
        else:
            self.thresholds = {
                "folder": {
                    "warn": 25,
                    "error": 50},
                "service": {
                    "warn": 50,
                    "error": 70}
            }

        # Read infrastructure login information
        if "login-file" in infra:
            logins = read_json(infra["login-file"])  # TODO: is this secure?
        else:
            self._log.warning("No login-file specified, defaulting to user prompts...")
            logins = {}

        # Instantiate the vSphere vCenter server instance class
        self.server = Vsphere(username=logins.get("user"),
                              password=logins.get("pass"),
                              hostname=infra.get("hostname"),
                              port=int(infra.get("port")),
                              datastore=infra.get("datastore"),
                              datacenter=infra.get("datacenter"))

        # Acquire ESXi hosts (TODO: make these their own classes /w network, etc methods)
        if "hosts" in infra:
            hosts = infra["hosts"]
            self.host = self.server.get_host(hosts[0])  # TODO: temporary hack
            self.hosts = [self.server.get_host(h) for h in hosts]  # Gather all the ESXi hosts
        else:
            self.host = self.server.get_host()  # First host found in Datacenter

        # Instantiate and initialize Groups
        self.groups = self._init_groups()

        # Set the server root folder (TODO: network folder in infrastructure spec)
        if "server-root" in infra:
            self.server_root = self.server.get_folder(infra["server-root"])
            if not self.server_root:
                self._log.error("Could not find server-root folder '%s'", infra["server-root"])
                exit(1)
        else:
            self.server_root = self.server.datacenter.vmFolder  # Default to Datacenter VM folder
        self._log.info("Server root folder: %s", self.server_root.name)

        # Set environment root folder (TODO: this can be simplified and consolidated)
        if "folder-name" not in self.metadata:
            self.root_path, self.root_name = ("", self.metadata["name"])
            self.root_folder = futils.traverse_path(folder=self.server_root,
                                                    path=self.root_name,
                                                    generate=True)
        else:
            self.root_path, self.root_name = os.path.split(self.metadata["folder-name"])
            self.root_folder = futils.traverse_path(folder=self.server_root,
                                                    path=self.metadata["folder-name"],
                                                    generate=True)

        self._log.debug("Environment root folder name: %s", self.root_name)
        if not self.root_folder:  # Create if it's not found
            parent = futils.traverse_path(self.server_root, self.root_path)
            self.root_folder = self.server.create_folder(self.root_name, parent)
            if not self.root_folder:
                self._log.error("Could not create root folder '%s'", self.root_name)
                exit(1)
        self._log.info("Environment root folder: %s", self.root_folder.name)

        # Set default vSwitch name
        if "vswitch" in infra:
            self.vswitch_name = infra["vswitch"]
        else:  # TODO: is this a good default? (If you have to ask this, it's probably not)
            from pyVmomi import vim
            self.vswitch_name = self.server.get_item(vim.Network).name

        self._log.debug("Finished initializing VsphereInterface")

    def _init_groups(self):
        """
        Instantiate and initialize Groups
        :return: Initialized Groups
        :rtype: dict(:class:`Group`)
        """
        from adles.group import Group, get_ad_groups
        groups = {}

        # Instantiate Groups
        for name, config in self.spec["groups"].items():
            if "instances" in config:  # Template groups
                groups[name] = [Group(name, config, i) for i in range(1, config["instances"] + 1)]
            else:  # Standard groups
                groups[name] = Group(name=name, group=config)

        # Initialize Active Directory-type Group user names
        ad_groups = get_ad_groups(groups)
        for g in ad_groups:
            # res = self.server.get_users(belong_to_group=g.ad_group, find_users=True) (TODO)
            res = None
            if res is not None:
                for r in res:
                    if r.group is True:  # Reference: pyvmomi/docs/vim/UserSearchResult.rst
                        self._log.error("Result '%s' is not a user", str(r))
                    else:
                        g.users.append(r.principal)
                g.size = (len(g.users) if len(g.users) > 1 else 1)  # Set the size, default to 1
            else:
                self._log.error("Could not initialize AD-group %s", str(g.ad_group))

        if hasattr(self.server.user_dir, "domainList"):
            self._log.debug("Domains on server: %s", str(self.server.user_dir.domainList))

        return groups

    def create_masters(self):
        """ Exercise Environment Master creation phase """

        # Get folder containing templates
        self.template_folder = futils.traverse_path(self.server_root, self.infra["template-folder"])
        if not self.template_folder:
            self._log.error("Could not find template folder in path '%s'",
                            self.infra["template-folder"])
            return
        else:
            self._log.debug("Found template folder: '%s'", self.template_folder.name)

        # Create master folder to hold base service instances
        self.master_folder = futils.traverse_path(self.root_folder, self.master_root_name)
        if not self.master_folder:
            self.master_folder = self.server.create_folder(self.master_root_name, self.root_folder)
            self._log.info("Created Master folder '%s' in '%s'",
                           self.master_root_name, self.root_name)

        # TODO: implement configuration of "network-interface" in the "services" top-level section
        # Create networks for master instances
        for net in self.networks:  # Iterate through the base types
            self._create_master_networks(net_type=net, default_create=True)

        # Create Master instances
        # TODO: Apply master-group permissions [default: group permissions]
        self._master_parent_folder_gen(self.folders, self.master_folder)

        # Output fully deployed master folder tree to debugging
        self._log.debug(futils.format_structure(futils.enumerate_folder(self.root_folder)))

    def _master_parent_folder_gen(self, folder, parent):
        """
        Generates parent-type Master folders
        :param dict folder: Dict with the folder tree structure as in spec
        :param parent: Parent folder
        :type parent: vim.Folder
        """
        skip_keys = ["instances", "description", "enabled"]
        if not self._is_enabled(folder):  # Check if disabled
            self._log.warning("Skipping disabled parent-type folder %s", parent.name)
            return

        group = None
        master_group = None

        # We have to check every item, as they could be either keywords or sub-folders
        for sub_name, sub_value in folder.items():
            if sub_name in skip_keys:  # Skip configurations that are not relevant
                continue
            elif sub_name == "group":
                group = self._get_group(sub_value)
            elif sub_name == "master-group":
                master_group = self._get_group(sub_value)
            else:
                folder_name = self.master_prefix + sub_name
                new_folder = self.server.create_folder(folder_name, create_in=parent)

                if "services" in sub_value:  # It's a base folder
                    if self._is_enabled(sub_value):
                        self._log.info("Generating Master base-type folder %s", sub_name)
                        self._master_base_folder_gen(sub_name, sub_value, new_folder)
                    else:
                        self._log.warning("Skipping disabled base-type folder %s", sub_name)
                else:  # It's a parent folder, recurse
                    if self._is_enabled(sub_value):
                        self._master_parent_folder_gen(sub_value, parent=new_folder)
                        self._log.info("Generating Master parent-type folder %s", sub_name)
                    else:
                        self._log.warning("Skipping disabled parent-type folder %s", sub_name)

        # TODO: apply master group permissions
        if master_group is None:
            master_group = group

    def _master_base_folder_gen(self, folder_name, folder_dict, parent):
        """
        Generates base-type Master folders
        :param str folder_name: Name of the base folder
        :param dict folder_dict: Dict with the base folder tree as in spec
        :param parent: Parent folder
        :type parent: vim.Folder
        """
        # Set the group to apply permissions for (TODO: apply permissions)
        if "master-group" in folder_dict:
            master_group = self._get_group(folder_dict["master-group"])
        else:
            master_group = self._get_group(folder_dict["group"])

        # Create Master instances
        for sname, sconfig in folder_dict["services"].items():
            if not self._is_vsphere(sconfig["service"]):
                self._log.debug("Skipping non-vsphere service '%s'", sname)
                continue

            self._log.info("Creating Master instance '%s' from service '%s'",
                           sname, sconfig["service"])

            vm = self._create_service(parent, sconfig["service"], sconfig["networks"])
            if vm is None:
                self._log.error("Failed to create Master instance '%s' in folder '%s'",
                                sname, folder_name)
                continue  # Skip to the next service

    def _create_service(self, folder, service_name, networks):
        """
        Retrieves and clones a service into a master folder
        :param folder: Folder to create service in
        :type folder: vim.Folder
        :param str service_name: Name of the service to clone
        :param list networks: Networks to configure the service with
        :return: The service VM instance
        :rtype: :class:`VM`
        """
        if not self._is_vsphere(service_name):
            self._log.debug("Skipping non-vsphere service '%s'", service_name)
            return None

        config = self.services[service_name]
        vm_name = self.master_prefix + service_name

        test = futils.traverse_path(folder, vm_name)  # Check service already exists
        if test is None:
            # Find the template that matches the service definition
            template = futils.traverse_path(self.template_folder, config["template"])
            if not template:
                self._log.error("Could not find template '%s' for service '%s'",
                                config["template"], service_name)
                return None
            self._log.info("Creating service '%s'", service_name)
            vm = VM(name=vm_name, folder=folder, resource_pool=self.server.get_pool(),
                    datastore=self.server.datastore, host=self.host)
            if not vm.create(template=template):
                return None
        else:
            self._log.warning("Service %s already exists", service_name)
            vm = VM(vm=test)
            # TODO: we're assuming the template's configuration is valid, need a way to check
            if vm.is_template():  # Check if it's been converted already
                self._log.warning("Service %s is a Template, skipping configuration", service_name)
                return vm

        if "resource-config" in config:  # Resource configurations (minus storage currently)
            vm.edit_resources(**config["resource-config"])

        if "note" in config:  # Set VM note if specified
            vm.set_note(config["note"])

        # NOTE: management interfaces matter here!
        self._configure_nics(vm, networks=networks)  # Configure VM NICs

        # Post-creation snapshot
        vm.create_snapshot("Start of Mastering",
                           "Beginning of Mastering phase for exercise %s", self.metadata["name"])
        return vm

    def _create_master_networks(self, net_type, default_create):
        """
        Creates a network as part of the Master creation phase
        :param str net_type: Top-level type of the network (unique | generic | base)
        :param bool default_create: Whether to create networks if they don't already exist
        """
        self.host.configManager.networkSystem.RefreshNetworkSystem()  # Pick up any recent changes
        self._log.info("Creating %s", net_type)

        for name, config in self.networks[net_type].items():
            exists = self.server.get_network(name)
            if exists:
                self._log.info("PortGroup '%s' already exists on host '%s'", name, self.host.name)
            else:  # NOTE: if monitoring, we want promiscuous=True
                self._log.warning("PortGroup '%s' does not exist on host '%s'",
                                  name, self.host.name)
                if default_create:
                    self._log.info("Creating portgroup '%s' on host '%s'", name, self.host.name)
                    create_portgroup(name=name, host=self.host, promiscuous=False,
                                     vlan=int(config.get("vlan", next(get_vlan()))),
                                     vswitch_name=config.get("vswitch", self.vswitch_name))

    def _configure_nics(self, vm, networks, instance=None):
        """
        Configures Virtual Network Interfaces Cards (vNICs) for a service instance
        :param vm: Virtual Machine to configure vNICs on
        :type vm: vim.VirtualMachine
        :param list networks: List of networks to configure
        :param int instance: Current instance of a folder for Deployment purposes [default: None]
        """
        self._log.info("Editing NICs for VM '%s'", vm.name)
        num_nics = len(list(vm.network))
        num_nets = len(networks)
        nets = networks  # Copy the passed variable so we can edit it later

        # Ensure number of NICs on VM matches number of networks configured for the service
        # Note that monitoring interfaces will be counted and included in the networks list
        if num_nics > num_nets:     # Remove excess interfaces
            diff = int(num_nics - num_nets)
            self._log.debug("VM '%s' has %d extra NICs, removing...", vm.name, diff)
            for _, nic in enumerate(reversed(range(num_nics)), start=1):
                vm.remove_nic(nic)
        elif num_nics < num_nets:   # Create missing interfaces
            diff = int(num_nets - num_nics)
            self._log.debug("VM '%s' is deficient %d NICs, adding...", vm.name, diff)
            for _ in range(diff):   # Add NICs to VM and pop them from the list of networks
                nic_model = ("vmxnet3" if vm.has_tools() else "e1000")  # Select NIC hardware
                net_name = nets.pop()
                vm.add_nic(network=self.server.get_network(net_name),
                           model=nic_model, summary=net_name)

        # Edit the interfaces (NOTE: any NICs added earlier shouldn't be affected by this)
        for i, net_name in enumerate(networks, start=1):
            # Setting the summary to network name allows viewing of name without requiring
            # read permissions to the network itself
            if instance is not None:  # Resolve generic networks for deployment phase
                net_name = self._get_net(net_name, instance)
            network = self.server.get_network(net_name)
            if vm.get_nic_by_id(i).backing.network == network:
                continue  # Skip NICs that are already configured
            else:
                vm.edit_nic(nic_id=i, network=network, summary=net_name)

    def deploy_environment(self):
        """ Exercise Environment deployment phase """
        self.master_folder = futils.traverse_path(self.root_folder, self.master_root_name)
        if self.master_folder is None:  # Check if Master folder was found
            self._log.error("Could not find Master folder '%s'. "
                            "Please ensure the  Master Creation phase has been run "
                            "and the folder exists before attempting Deployment",
                            self.master_root_name)
            exit(1)
        self._log.debug("Master folder name: %s\tPrefix: %s",
                        self.master_folder.name, self.master_prefix)

        # Verify and convert Master instances to templates
        self._log.info("Validating and converting Masters to Templates")
        self._convert_and_verify(folder=self.master_folder)
        self._log.info("Finished validating and converting Masters to Templates")

        self._log.info("Deploying environment...")
        self._deploy_parent_folder_gen(spec=self.folders, parent=self.root_folder, path="")
        self._log.info("Finished deploying environment")

        # Output fully deployed environment tree to debugging
        self._log.debug(futils.format_structure(futils.enumerate_folder(self.root_folder)))

    def _convert_and_verify(self, folder):
        """
        Converts Masters to Templates before deployment.
        This also ensures they are powered off before being cloned.
        :param folder: Folder containing Master instances to convert and verify
        :type folder: vim.Folder
        """
        self._log.debug("Converting Masters in folder '%s' to templates", folder.name)
        for item in folder.childEntity:
            if is_vm(item):
                vm = VM(vm=item)
                self.masters[vm.name] = vm
                if vm.is_template():  # Skip if they already exist from a previous run
                    self._log.debug("Master '%s' is already a template", vm.name)
                    continue

                # Cleanly power off VM before converting to template
                if vm.powered_on():
                    vm.change_state("off", attempt_guest=True)

                # Take a snapshot to allow reverts to the start of the exercise
                vm.create_snapshot("Start of exercise",
                                   "Beginning of deployment phase, post-master configuration")

                # Convert Master instance to Template
                vm.convert_template()
                if not vm.is_template():
                    self._log.error("Master '%s' did not convert to Template", vm.name)
                else:
                    self._log.debug("Converted Master '%s' to Template", vm.name)
            elif is_folder(item):  # Recurse into sub-folders
                self._convert_and_verify(item)
            else:
                self._log.debug("Unknown item found while templatizing Masters: %s", str(item))

    def _deploy_parent_folder_gen(self, spec, parent, path):
        """
        Generates parent-type folder trees
        :param dict spec: Dict with folder specification
        :param parent: Parent folder
        :type parent: vim.Folder
        :param str path: Folders path at the current level
        """
        skip_keys = ["instances", "description", "master-group", "enabled"]
        if not self._is_enabled(spec):  # Check if disabled
            self._log.warning("Skipping disabled parent-type folder %s", parent.name)
            return

        for sub_name, sub_value in spec.items():
            if sub_name in skip_keys:  # Skip configurations that are not relevant
                continue
            elif sub_name == "group":  # Configure group
                group = self._get_group(sub_value)  # TODO: apply group permissions
            else:  # Create instances of the parent folder
                self._log.debug("Deploying parent-type folder '%s'", sub_name)
                num_instances, prefix = self._instances_handler(spec, sub_name, "folder")
                for i in range(num_instances):
                    # If prefix is undefined or there's a single instance, use the folder's name
                    instance_name = (sub_name if prefix == "" or num_instances == 1 else prefix)

                    # If multiple instances, append padded instance number
                    instance_name += (pad(i) if num_instances > 1 else "")

                    # Create a folder for the instance
                    new_folder = self.server.create_folder(instance_name, create_in=parent)

                    if "services" in sub_value:  # It's a base folder
                        if self._is_enabled(sub_value):
                            self._deploy_base_folder_gen(folder_name=sub_name,
                                                         folder_items=sub_value, parent=new_folder,
                                                         path=self._path(path, sub_name))
                        else:
                            self._log.warning("Skipping disabled base-type folder %s", sub_name)
                    else:  # It's a parent folder
                        if self._is_enabled(sub_value):
                            self._deploy_parent_folder_gen(parent=new_folder, spec=sub_value,
                                                           path=self._path(path, sub_name))
                        else:
                            self._log.warning("Skipping disabled parent-type folder %s", sub_name)

    def _deploy_base_folder_gen(self, folder_name, folder_items, parent, path):
        """
        Generates folder tree for deployment stage
        :param str folder_name: Name of the folder
        :param dict folder_items: Dict of items in the folder
        :param parent: Parent folder
        :type parent: vim.Folder
        :param str path: Folders path at the current level
        """
        # Set the group to apply permissions for (TODO: apply permissions)
        group = self._get_group(folder_items["group"])

        # Get number of instances and check if it exceeds configured limits
        num_instances, prefix = self._instances_handler(folder_items, folder_name, "folder")

        # Create instances
        self._log.info("Deploying base-type folder '%s'", folder_name)
        for i in range(num_instances):
            # If no prefix is defined or there's only a single instance, use the folder's name
            instance_name = (folder_name if prefix == "" or num_instances == 1 else prefix)

            # If multiple instances, append padded instance number
            instance_name += (pad(i) if num_instances > 1 else "")

            if num_instances > 1:  # Create a folder for the instance
                new_folder = self.server.create_folder(instance_name, create_in=parent)
            else:  # Don't duplicate folder name for single instances
                new_folder = parent

            # Use the folder's name for the path, as that's what matches the Master version
            self._log.info("Generating services for base-type folder instance '%s'", instance_name)
            self._deploy_gen_services(services=folder_items["services"], parent=new_folder,
                                      path=path, instance=i)

    def _deploy_gen_services(self, services, parent, path, instance):
        """
        Generates the services in a folder
        :param dict services: The "services" dict in a folder
        :param parent: Parent folder
        :type parent: vim.Folder
        :param str path: Folders path at the current level
        :param int instance: What instance of a base folder this is
        """
        # Iterate through the services
        for service_name, value in services.items():
            if not self._is_vsphere(value["service"]):  # Ignore non-vsphere services
                self._log.debug("Skipping non-vsphere service '%s'", service_name)
                continue
            self._log.info("Generating service '%s' in folder '%s'", service_name, parent.name)

            # Get number of instances for the service and check if it exceeds configured limits
            num_instances, prefix = self._instances_handler(value, service_name, "service")

            # Get the Master template instance to clone from
            master = self.masters.get(self.master_prefix + value["service"], None)
            if master is None:  # Check if the lookup was successful
                self._log.error("Couldn't find Master for service '%s' in this path:\n%s",
                                value["service"], path)
                continue  # Skip to the next service

            # Clone the instances of the service from the master
            for i in range(num_instances):
                instance_name = prefix + service_name + (" " + pad(i) if num_instances > 1 else "")
                vm = VM(name=instance_name, folder=parent, resource_pool=self.server.get_pool(),
                        datastore=self.server.datastore, host=self.host)
                if not vm.create(template=master.get_vim_vm()):
                    self._log.error("Failed to create instance %s", instance_name)
                else:
                    self._configure_nics(vm, value["networks"], instance=instance)

    def _is_vsphere(self, service_name):
        """
        Checks if a service instance is defined as a vSphere service
        :param str service_name: Name of the service to lookup in list of defined services
        :return: If a service is a vSphere-type service
        :rtype: bool
        """
        # TODO: make "template" and other platform identifiers global keywords
        if service_name not in self.services:
            self._log.error("Could not find service %s in list of services", service_name)
        elif "template" in self.services[service_name]:
            return True
        return False

    def _get_net(self, name, instance=-1):
        """
        Resolves network names. This is mainly to handle generic-type networks.
        If a generic network does not exist, it is created and added to the interface lookup table.
        :param str name: Name of the network
        :param int instance: Instance number (Only applies to generic-type networks)
        :return: Resolved network name
        :rtype: str
        """
        # TODO: could use this to do network lookups on the server as well
        net_type = self._determine_net_type(name)
        if net_type == "unique-networks":
            return name
        elif net_type == "generic-networks":
            if instance == -1:
                self._log.error("Invalid instance for _get_net: %d", instance)
                raise ValueError
            net_name = name + "-GENERIC-" + pad(instance)  # Generate full name for the generic net
            if net_name not in self.net_table:
                exists = self.server.get_network(net_name)
                if exists is not None:
                    self._log.debug("PortGroup '%s' already exists on host '%s'", net_name,
                                    self.host.name)
                else:  # Create the generic network if it does not exist
                    # NOTE: lookup of name is case-sensitive! This can lead to bugs
                    self._log.debug("Creating portgroup '%s' on host '%s'", net_name,
                                    self.host.name)
                    vsw = self.networks["generic-networks"][name].get("vswitch", self.vswitch_name)
                    create_portgroup(name=net_name, host=self.host, promiscuous=False,
                                     vlan=next(get_vlan()), vswitch_name=vsw)
                self.net_table[net_name] = True  # Register the existence of the generic
            return net_name
        else:
            self._log.error("Invalid network type %s for network %s", net_type, name)
            raise TypeError

    def cleanup_masters(self, network_cleanup=False):
        """
        Cleans up any master instances
        :param bool network_cleanup: If networks should be cleaned up [default: False]
        """
        # TODO: look at getorphanedvms in pyvmomi-community-samples for how to do this
        # Get the folder to cleanup in
        master_folder = futils.find_in_folder(self.root_folder, self.master_root_name)
        self._log.info("Found master folder '%s' under folder '%s', proceeding with cleanup...",
                       master_folder.name, self.root_folder.name)

        # Recursively descend from master folder, destroying anything with the prefix
        futils.cleanup(folder=master_folder, vm_prefix=self.master_prefix,
                       recursive=True, destroy_folders=True, destroy_self=True)

        # Cleanup networks (TODO: use network folders to aid in this, during creation phase)
        if network_cleanup:
            pass

    def cleanup_environment(self, network_cleanup=False):
        """
        Cleans up a deployed environment
        :param bool network_cleanup: If networks should be cleaned up [default: False]
        """
        # Get the root environment folder to cleanup in
        enviro_folder = self.root_folder

        # TODO: ensure master folder is skipped

        # Cleanup networks (TODO: use network folders to aid in this, during creation phase)
        if network_cleanup:
            pass

    def __str__(self):
        return str(self.server) + str(self.groups) + str(self.hosts)

    def __eq__(self, other):
        return super(VsphereInterface, self).__eq__(other) and self.server == other.server and \
               self.groups == other.groups and self.hosts == other.hosts



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from adles.utils import time_execution


class Interface:
    """ Generic interface used to uniformly interact with platform-specific interfaces. """
    __version__ = "1.1.1"

    # Names/prefixes
    master_prefix = "(MASTER) "
    master_root_name = "MASTER-FOLDERS"

    def __init__(self, infra, spec):
        """ 
        :param dict infra: Full infrastructure configuration
        :param dict spec: Full exercise specification
        """
        self._log = logging.getLogger('PlatformInterface')
        self._log.debug("Initializing Interface %s", Interface.__version__)

        self.infra = infra      # Save the infrastructure configuration
        self.spec = spec        # Save the exercise specification
        self.metadata = spec["metadata"]    # Save the exercise spec metadata
        self.networks = spec["networks"]    # Networks for platforms
        self.interfaces = []    # List of instantiated platform interfaces
        self.thresholds = {}    # Thresholds for platforms
        self.groups = {}        # Groups for platforms

        # Select the Interface to use based on the specified infrastructure platform
        for platform, config in infra.items():
            if platform == "vmware-vsphere":
                # TODO: support multiple vSphere instances (e.g for remote labs)
                from .vsphere_interface import VsphereInterface
                self.interfaces.append(VsphereInterface(config, spec))
            elif platform == "docker":
                from .docker_interface import DockerInterface
                self.interfaces.append(DockerInterface(config, spec))
            else:
                self._log.error("Invalid platform: %s", str(platform))
                raise ValueError

    # TODO: subprocess each interface call
    @time_execution
    def create_masters(self):
        """ Master creation phase """
        self._log.info("Creating Master instances for %s", self.metadata["name"])
        for i in self.interfaces:
            i.create_masters()

    @time_execution
    def deploy_environment(self):
        """ Environment deployment phase """
        self._log.info("Deploying environment for %s", self.metadata["name"])
        for i in self.interfaces:
            i.deploy_environment()

    @time_execution
    def cleanup_masters(self, network_cleanup=False):
        """
        Cleans up master instances
        :param bool network_cleanup: If networks should be cleaned up [default: False]
        """
        self._log.info("Cleaning up Master instances for %s", self.metadata["name"])
        for i in self.interfaces:
            i.cleanup_masters(network_cleanup=network_cleanup)

    @time_execution
    def cleanup_environment(self, network_cleanup=False):
        """
        Cleans up a deployed environment
        :param bool network_cleanup: If networks should be cleaned up [default: False]
        """
        self._log.info("Cleaning up environment for %s", self.metadata["name"])
        for i in self.interfaces:
            i.cleanup_masters(network_cleanup=network_cleanup)

    def _instances_handler(self, spec, obj_name, obj_type):
        """
        Determines number of instances and optional prefix using specification
        :param dict spec: Dict of folder
        :param str obj_name: Name of the thing being handled
        :param str obj_type: Type of the thing being handled (folder | service)
        :return: Number of instances, Prefix
        :rtype: tuple(int, str)
        """
        num = 1
        prefix = ""
        if "instances" in spec:
            if type(spec["instances"]) == int:
                num = int(spec["instances"])
            else:
                if "prefix" in spec["instances"]:
                    prefix = str(spec["instances"]["prefix"])
                if "number" in spec["instances"]:
                    num = int(spec["instances"]["number"])
                elif "size-of" in spec["instances"]:
                    # size_of = spec["instances"]["size-of"])
                    # num = int(self._get_group(size_of).size
                    # if num < 1:
                    num = 1  # TODO: WORKAROUND FOR AD-GROUPS
                else:
                    self._log.error("Unknown instances specification: %s", str(spec["instances"]))
                    num = 0

        # Check if the number of instances exceeds the configured thresholds for the interface
        thr = self.thresholds[obj_type]
        if num > thr["error"]:
            self._log.error("%d instances of %s '%s' is beyond the configured %s threshold of %d",
                            num, obj_type, obj_name, self.__name__, thr["error"])
            raise Exception("Threshold exceeded")
        elif num > thr["warn"]:
            self._log.warning("%d instances of %s '%s' is beyond the configured %s threshold of %d",
                              num, obj_type, obj_name, self.__name__, thr["warn"])
        return num, prefix

    def _path(self, path, name):
        """
        Generates next step of the path for deployment of Masters
        :param str path: Current path
        :param str name: Name to add to the path
        :return: The updated path
        :rtype: str
        """
        return str(path + '/' + self.master_prefix + name)

    @staticmethod
    def _is_enabled(spec):
        """
        Determines if a spec is enabled
        :param dict spec: 
        :return: If the spec is enabled
        :rtype: bool
        """
        if "enabled" in spec:
            return bool(spec["enabled"])
        else:
            return True

    def _determine_net_type(self, network_label):
        """
        Determines the type of a network
        :param str network_label: Name of network to determine type of
        :return: Type of the network ("generic-networks" | "unique-networks")
        :rtype: str
        """
        for net_name, net_value in self.networks.items():
            vals = set(k for k in net_value)
            if network_label in vals:
                return net_name
        self._log.error("Could not find type for network '%s'", network_label)
        return ""

    def _get_group(self, group_name):
        """
        Provides a uniform way to get information about normal groups and template groups
        :param str group_name: Name of the group
        :return: Group object
        :rtype: :class:`Group`
        """
        from adles.group import Group
        if group_name in self.groups:
            g = self.groups[group_name]
            if isinstance(g, Group):    # Normal groups
                return g
            elif isinstance(g, list):   # Template groups
                return g[0]
            else:
                self._log.error("Unknown type for group '%s': %s", str(group_name), str(type(g)))
        else:
            self._log.error("Could not get group '%s' from groups", group_name)

    def __repr__(self):
        return "%s(%s, %s)" % (str(self.__class__), str(self.spec), str(self.infra))

    def __str__(self):
        return str([x for x in self.infra.keys()])

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.infra == other.infra and \
               self.spec == other.spec

    def __ne__(self, other):
        return not self.__eq__(other)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from pyVmomi import vim

import adles.utils as utils
from adles.vsphere.folder_utils import find_in_folder


# Docs: http://pubs.vmware.com/vsphere-60/topic/com.vmware.wssdk.apiref.doc/vim.VirtualMachine.html
class VM:
    """ Represents a VMware vSphere Virtual Machine instance.
     
    .. warning::    You must call :meth:`create` if a vim.VirtualMachine object is 
                    not used to initialize the instance.
    """
    __version__ = "0.6.0"

    def __init__(self, vm=None, name=None, folder=None, resource_pool=None,
                 datastore=None, host=None):
        """
        :param vm: VM instance to use instead of calling :meth:`create`
        :type vm: vim.VirtualMachine
        :param str name: Name of the VM
        :param folder: Folder in inventory to create the VM in
        :type folder: vim.Folder
        :param resource_pool: Resource pool to use for the VM
        :type resource_pool: vim.ResourcePool
        :param datastore: Datastore the VM is stored on
        :type datastore: vim.Datastore
        :param host: Host the VM runs on
        :type host: vim.HostSystem
        """
        self._log = logging.getLogger('VM')
        if vm is not None:
            self._vm = vm
            self.name = vm.name
            self.folder = vm.parent
            self.resource_pool = vm.resourcePool
            self.datastore = vm.datastore[0]
            self.host = vm.summary.runtime.host
            self.network = vm.network
            self.runtime = vm.runtime
            self.summary = vm.summary
        else:
            self._vm = None
            self.name = name
            self.folder = folder  # vim.Folder that will contain the VM
            self.resource_pool = resource_pool  # vim.ResourcePool to use for the VM
            self.datastore = datastore  # vim.Datastore object to store VM on
            self.host = host  # vim.HostSystem (TODO: may be Host class soon)

    @utils.time_execution
    def create(self, template=None, cpus=1, cores=1, memory=512, max_consoles=None,
               version=None, firmware='efi', datastore_path=None):
        """
        Creates a Virtual Machine
        :param template: Template VM to clone
        :type template: vim.VirtualMachine
        :param int cpus: Number of processors [default: 1]
        :param int cores: Number of processor cores [default: 1]
        :param int memory: Amount of RAM in MB [default: 512]
        :param int max_consoles: Maximum number of active console connections [default: default]
        :param int version: Hardware version of the VM [default: highest host supports]
        :param str firmware: Firmware to emulate for the VM (efi | bios) [default: efi]
        :param str datastore_path: Path to existing VM files on datastore [default: None]
        :return: If the creation was successful
        :rtype: bool
        """
        if template is not None:  # Use a template to create the VM
            self._log.debug("Creating VM '%s' by cloning %s", self.name, template.name)
            clonespec = vim.vm.CloneSpec()
            clonespec.location = vim.vm.RelocateSpec(pool=self.resource_pool,
                                                     datastore=self.datastore)
            if not template.CloneVM_Task(folder=self.folder, name=self.name,
                                         spec=clonespec).wait(120):
                self._log.error("Error cloning VM %s", self.name)
                return False
            # TODO: if cloned, reconfigure to match anything given as parameters, e.g memory
        else:  # Generate the specification for and create the new VM
            self._log.debug("Creating VM '%s' from scratch", self.name)
            spec = vim.vm.ConfigSpec()
            spec.name = self.name
            spec.numCPUs = int(cpus)
            spec.numCoresPerSocket = int(cores)
            spec.cpuHotAddEnabled = True
            spec.memoryMB = int(memory)
            spec.memoryHotAddEnabled = True
            spec.firmware = str(firmware).lower()
            if version is not None:
                spec.version = str(version)
            if max_consoles is not None:
                spec.maxMksConnections = int(max_consoles)
            vm_path = '[' + self.datastore.name + '] '
            if datastore_path:
                vm_path += str(datastore_path)
            vm_path += self.name + '/' + self.name + '.vmx'
            spec.files = vim.vm.FileInfo(vmPathName=vm_path)
            self._log.debug("Creating VM '%s' in folder '%s'", self.name, self.folder.name)
            if not self.folder.CreateVM_Task(spec, self.resource_pool, self.host).wait():
                self._log.error("Error creating VM %s", self.name)
                return False

        self._vm = find_in_folder(self.folder, self.name, vimtype=vim.VirtualMachine)
        if not self._vm:
            self._log.error("Failed to make VM %s", self.name)
            return False
        else:
            self._log.debug("Created VM %s", self.name)
        self.network = self._vm.network
        self.runtime = self._vm.runtime
        self.summary = self._vm.summary
        return True

    def destroy(self):
        """ Destroys the VM """
        self._log.debug("Destroying VM %s", self.name)
        if self.powered_on():
            self.change_state("off")
        self._vm.Destroy_Task().wait()

    def change_state(self, state, attempt_guest=True):
        """
        Generic power state change function that uses guest operations if available
        :param str state: State to change to (on | off | reset | suspend)
        :param bool attempt_guest: Whether to attempt to use guest operations to change power state
        """
        state = state.lower()  # Convert to lowercase for comparisons
        if self.is_template():
            self._log.error("VM '%s' is a Template, so state cannot be changed to '%s'",
                            self.name, state)
        elif attempt_guest and self.has_tools() and state != "on":  # Can't power on using guest ops
            if self._vm.summary.guest.toolsStatus == "toolsNotInstalled":
                self._log.error("Cannot change a VM's guest power state without VMware Tools!")
                return
            elif state == "shutdown" or state == "off":
                task = self._vm.ShutdownGuest()
            elif state == "reboot" or state == "reset":
                task = self._vm.RebootGuest()
            elif state == "standby" or state == "suspend":
                task = self._vm.StandbyGuest()
            else:
                self._log.error("Invalid guest_state argument: %s", state)
                return
            self._log.debug("Changing guest power state of VM %s to: '%s'", self.name, state)
            try:
                task.wait()
            except vim.fault.ToolsUnavailable:
                self._log.error("Can't change guest state of '%s': Tools aren't running", self.name)
        else:
            if state == "on":
                task = self._vm.PowerOnVM_Task()
            elif state == "off":
                task = self._vm.PowerOffVM_Task()
            elif state == "reset":
                task = self._vm.ResetVM_Task()
            elif state == "suspend":
                task = self._vm.SuspendVM_Task()
            else:
                self._log.error("Invalid state arg %s for VM %s", state, self.name)
                return
            self._log.debug("Changing power state of VM %s to: '%s'", self.name, state)
            task.wait()

    def edit_resources(self, cpus=None, cores=None, memory=None, max_consoles=None):
        """
        Edits resource limits for the VM
        :param int cpus: Number of CPUs
        :param int cores: Number of CPU cores
        :param int memory: Amount of RAM in MB
        :param int max_consoles: Maximum number of simultaneous MKS console connections
        """
        spec = vim.vm.ConfigSpec()
        if cpus is not None:
            spec.numCPUs = int(cpus)
        if cores is not None:
            spec.numCoresPerSocket = int(cores)
        if memory is not None:
            spec.memoryMB = int(memory)
        if max_consoles is not None:
            spec.maxMksConnections = int(max_consoles)
        self._edit(spec)

    def rename(self, name):
        """
        Renames the VM
        :param str name: New name for the VM
        """
        self._log.debug("Renaming VM %s to %s", self.name, name)
        if not self._vm.Rename_Task(newName=str(name)).wait():
            self._log.error("Failed to rename VM %s to %s", self.name, name)
        else:
            self.name = str(name)

    def upgrade(self, version):
        """
        Upgrades the hardware version of the VM
        :param int version: Version of hardware to upgrade VM to [default: latest host supports]
        """
        full_version = "vmx-" + str(version)
        try:
            self._vm.UpgradeVM_Task(full_version).wait()
        except vim.fault.AlreadyUpgraded:
            self._log.warning("Hardware version is already up-to-date for %s", self.name)

    def convert_template(self):
        """ Converts a Virtual Machine to a Template """
        if self.is_template():
            self._log.warning("%s is already a Template", self.name)
        else:
            self._log.debug("Converting '%s' to Template", self.name)
            self._vm.MarkAsTemplate()

    def convert_vm(self):
        """ Converts a Template to a Virtual Machine """
        self._log.debug("Converting '%s' to VM", self.name)
        self._vm.MarkAsVirtualMachine(self.resource_pool, self.host)

    def set_note(self, note):
        """
        Sets the note on the VM
        :param str note: String to set the note to
        """
        self._edit(vim.vm.ConfigSpec(annotation=str(note)))

    # From: execute_program_in_vm.py in pyvmomi_community_samples
    def execute_program(self, process_manager, program_path,
                        username=None, password=None, program_args=""):
        """
        Executes a commandline program in the VM. 
        This requires VMware Tools to be installed on the VM.
        :param process_manager: vSphere process manager object
        :type process_manager: vim.vm.guest.ProcessManager
        :param str program_path: Path to the program inside the VM
        :param str username: User on VM to execute program using [default: current ADLES user]
        :param str password: Plaintext password for the User [default: prompt user]
        :param str program_args: Commandline arguments for the program [default: ""]
        :return: Program Process ID (PID) if it was executed successfully, -1 if not
        :rtype: int
        """
        # TODO: listprocesses/terminate process in guest, make helper func for guest authentication
        from os.path import basename
        prog_name = basename(program_path)
        if not self.has_tools():
            self._log.error("Cannot execute program %s in VM %s: VMware Tools is not running",
                            prog_name, self.name)
            return -1
        if username is None:
            from getpass import getuser
            username = getuser()
        if password is None:
            from getpass import getpass
            password = getpass("Enter password of user %s to execute program %s on VM %s"
                               % (username, prog_name, self.name))
        creds = vim.vm.guest.NamePasswordAuthentication(username=username, password=password)
        try:
            ps = vim.vm.guest.ProcessManager.ProgramSpec(programPath=program_path,
                                                         arguments=program_args)
            pid = process_manager.StartProgramInGuest(self._vm, creds, ps)
            self._log.debug("Successfully started program %s in VM %s, PID is %s",
                            prog_name, self.name, pid)
            return pid
        except IOError as e:
            self._log.error("Could not execute program %s in VM %s: %s",
                            prog_name, self.name, str(e))
            return -1

    def snapshot_disk_usage(self):
        """
        Determines the total disk usage of a VM's snapshots
        :return: Human-readable disk usage of the snapshots
        :rtype: str
        """
        from re import search
        disk_list = self._vm.layoutEx.file
        size = 0
        for disk in disk_list:
            if disk.type == 'snapshotData':
                size += disk.size
            ss_disk = search('0000\d\d', disk.name)
            if ss_disk:
                size += disk.size
        return utils.sizeof_fmt(size)

    def create_snapshot(self, name, description='', memory=False, quiesce=True):
        """
        Creates a snapshot of the VM
        :param str name: Name of the snapshot
        :param str description: Text description of the snapshot [default: '']
        :param bool memory: Memory dump of the VM is included in the snapshot [default: False]
        :param bool quiesce: Quiesce VM disks (Requires VMware Tools) [default: True]
        """
        self._log.info("Creating snapshot '%s' of VM '%s'", name, self.name)
        if not self._vm.CreateSnapshot_Task(name=name, description=description,
                                     memory=bool(memory), quiesce=quiesce).wait():
            self._log.error("Failed to take snapshot of VM %s", self.name)

    def revert_to_snapshot(self, snapshot):
        """
        Reverts VM to the named snapshot
        :param str snapshot: Name of the snapshot to revert to
        """
        self._log.info("Reverting '%s' to the snapshot '%s'", self.name, snapshot)
        self.get_snapshot(snapshot).RevertToSnapshot_Task().wait()

    def revert_to_current_snapshot(self):
        """ Reverts the VM to the most recent snapshot """
        self._log.info("Reverting '%s' to the current snapshot", self.name)
        self._vm.RevertToCurrentSnapshot_Task().wait()

    def remove_snapshot(self, snapshot, remove_children=True, consolidate_disks=True):
        """
        Removes the named snapshot from the VM
        :param str snapshot: Name of the snapshot to remove
        :param bool remove_children: Removal of the entire snapshot subtree [default: True]
        :param bool consolidate_disks: Virtual disks of deleted snapshot will be merged with
        other disks if possible [default: True]
        """
        self._log.info("Removing snapshot '%s' from '%s'", snapshot, self.name)
        self.get_snapshot(snapshot).RemoveSnapshot_Task(remove_children, consolidate_disks).wait()

    def remove_all_snapshots(self, consolidate_disks=True):
        """
        Removes all snapshots associated with the VM
        :param bool consolidate_disks: Virtual disks of the deleted snapshot will be merged with
        other disks if possible [default: True]
        """
        self._log.info("Removing ALL snapshots for %s", self.name)
        self._vm.RemoveAllSnapshots_Task(consolidate_disks).wait()

    # Originally based on: add_nic_to_vm in pyvmomi-community-samples
    def add_nic(self, network, summary="default-summary", model="e1000"):
        """
        Add a NIC in the portgroup to the VM
        :param network: Network to attach NIC to
        :type network: vim.Network
        :param str summary: Human-readable device info [default: default-summary]
        :param str model: Model of virtual network adapter. [default: e1000]
        Options: (e1000 | e1000e | vmxnet | vmxnet2 | vmxnet3 | pcnet32 | sriov)
        e1000 will work on Windows Server 2003+, and e1000e is supported on Windows Server 2012+.
        VMXNET adapters require VMware Tools to be installed, and provide enhanced performance.
        Read this for more details: http://rickardnobel.se/vmxnet3-vs-e1000e-and-e1000-part-1/
        """
        if not isinstance(network, vim.Network):
            self._log.error("Invalid network type when adding vNIC to VM '%s': %s",
                            self.name, type(network).__name__)
        self._log.debug("Adding NIC to VM '%s'\nNetwork: '%s'\tSummary: '%s'\tNIC Model: '%s'",
                        self.name, network.name, summary, model)
        nic_spec = vim.vm.device.VirtualDeviceSpec()  # Create base object to add configurations to
        nic_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.add

        # Set the type of network adapter
        if model == "e1000":
            nic_spec.device = vim.vm.device.VirtualE1000()
        elif model == "e1000e":
            nic_spec.device = vim.vm.device.VirtualE1000e()
        elif model == "vmxnet":
            nic_spec.device = vim.vm.device.VirtualVmxnet()
        elif model == "vmxnet2":
            nic_spec.device = vim.vm.device.VirtualVmxnet2()
        elif model == "vmxnet3":
            nic_spec.device = vim.vm.device.VirtualVmxnet3()
        elif model == "pcnet32":
            nic_spec.device = vim.vm.device.VirtualPCNet32()
        elif model == "sriov":
            nic_spec.device = vim.vm.device.VirtualSriovEthernetCard()
        else:
            self._log.error("Invalid NIC model: '%s'\nDefaulting to e1000...", model)
            nic_spec.device = vim.vm.device.VirtualE1000()
        nic_spec.device.addressType = 'generated'  # Sets how MAC address is assigned
        nic_spec.device.wakeOnLanEnabled = False  # Disables Wake-on-lan capabilities

        nic_spec.device.deviceInfo = vim.Description()
        nic_spec.device.deviceInfo.summary = summary

        nic_spec.device.backing = vim.vm.device.VirtualEthernetCard.NetworkBackingInfo()
        nic_spec.device.backing.useAutoDetect = False
        nic_spec.device.backing.network = network  # Sets port group to assign adapter to
        nic_spec.device.backing.deviceName = network.name  # Sets name of device on host system

        nic_spec.device.connectable = vim.vm.device.VirtualDevice.ConnectInfo()
        nic_spec.device.connectable.startConnected = True  # Ensures adapter is connected at boot
        nic_spec.device.connectable.allowGuestControl = True  # Allows guest OS to control device
        nic_spec.device.connectable.connected = True
        nic_spec.device.connectable.status = 'untried'
        # TODO: configure guest IP address if statically assigned
        self._edit(vim.vm.ConfigSpec(deviceChange=[nic_spec]))  # Apply change to VM

    def edit_nic(self, nic_id, network=None, summary=None):
        """
        Edits a vNIC based on it's number
        :param int nic_id: Number of network adapter on VM
        :param network: Network to assign the vNIC to [default: None]
        :type network: vim.Network
        :param str summary: Human-readable device description [default: None]
        """
        nic_label = 'Network adapter ' + str(nic_id)
        self._log.debug("Changing '%s' on VM '%s'", nic_label, self.name)
        virtual_nic_device = self.get_nic_by_name(nic_label)
        if not virtual_nic_device:
            self._log.error('Virtual %s could not be found!', nic_label)
            return
        nic_spec = vim.vm.device.VirtualDeviceSpec()
        nic_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.edit
        nic_spec.device = virtual_nic_device
        if summary:
            nic_spec.device.deviceInfo.summary = str(summary)
        if network:
            self._log.debug("Changing PortGroup to: '%s'", network.name)
            nic_spec.device.backing.network = network
            nic_spec.device.backing.deviceName = network.name
        self._edit(vim.vm.ConfigSpec(deviceChange=[nic_spec]))  # Apply change to VM

    # Originally based on: delete_nic_from_vm in pyvmomi-community-samples
    def remove_nic(self, nic_number):
        """
        Deletes a vNIC based on it's number
        :param int nic_number: Number of the vNIC to delete
        :return: If removal succeeded
        :rtype: bool
        """
        nic_label = "Network adapter " + str(nic_number)
        self._log.debug("Removing Virtual %s from '%s'", nic_label, self.name)
        virtual_nic_device = self.get_nic_by_name(nic_label)
        if virtual_nic_device is not None:
            virtual_nic_spec = vim.vm.device.VirtualDeviceSpec()
            virtual_nic_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.remove
            virtual_nic_spec.device = virtual_nic_device
            self._edit(vim.vm.ConfigSpec(deviceChange=[virtual_nic_spec]))  # Apply change to VM
            return True
        else:
            self._log.error("Virtual %s could not be found for '%s'", nic_label, self.name)
            return False

    def remove_device(self, device_spec):
        """
        Removes a device from the VM
        :param device_spec: The specification of the device to remove
        :type device_spec: vim.vm.device.VirtualDeviceSpec
        """
        self._log.debug("Removing device '%s' from '%s'", device_spec.name, self.name)
        device_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.remove
        self._edit(vim.vm.ConfigSpec(deviceChange=[device_spec]))

    # From: delete_disk_from_vm.py in pyvmomi_community_samples
    def remove_hdd(self, disk_number):
        """
        Removes a numbered Virtual Hard Disk from the VM
        :param int disk_number: Number of the HDD to remove
        :return: If the HDD was successfully removed
        :rtype: bool
        """
        hdd_label = "Hard disk " + str(disk_number)
        self._log.debug("Removing Virtual HDD %s from %s", hdd_label, self.name)
        dev = self.get_hdd_by_name(hdd_label)
        if not dev:
            self._log.error("Could not find Virtual %s to remove", hdd_label)
            return False
        else:
            spec = vim.vm.device.VirtualDeviceSpec(device=dev)
            self.remove_device(device_spec=spec)
            return True

    def attach_iso(self, iso_path, datastore=None, boot=True):
        """
        Attaches an ISO image to a VM
        :param str iso_path: Path in the Datastore of the ISO image to attach
        :param datastore: Datastore where the ISO resides [default: VM's datastore]
        :type datastore: vim.Datastore
        :param bool boot: Set VM to boot from the attached ISO [default: True]
        """
        self._log.debug("Adding ISO '%s' to '%s'", iso_path, self.name)
        if datastore is None:
            datastore = self.datastore

        drive_spec = vim.vm.device.VirtualDeviceSpec()
        drive_spec.device = vim.vm.device.VirtualCdrom()
        drive_spec.device.key = -1
        drive_spec.device.unitNumber = 0

        # Find a disk controller to attach to
        controller = self._find_free_ide_controller()
        if controller:
            drive_spec.device.controllerKey = controller.key
        else:
            self._log.error("Could not find a free IDE controller on '%s' to attach ISO '%s'",
                            self.name, iso_path)
            return

        drive_spec.device.backing = vim.vm.device.VirtualCdrom.IsoBackingInfo()
        drive_spec.device.backing.fileName = "[%s] %s" % (datastore.name, iso_path)  # Attach ISO
        drive_spec.device.backing.datastore = datastore  # Set datastore ISO is in

        drive_spec.device.connectable = vim.vm.device.VirtualDevice.ConnectInfo()
        drive_spec.device.connectable.allowGuestControl = True  # Allows guest OS to control device
        drive_spec.device.connectable.startConnected = True  # Ensures ISO is connected at boot

        drive_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.add
        vm_spec = vim.vm.ConfigSpec(deviceChange=[drive_spec])
        if boot:  # Set the VM to boot from the ISO upon power on
            self._log.debug("Setting '%s' to boot from ISO '%s'", self.name, iso_path)
            order = [vim.vm.BootOptions.BootableCdromDevice()]
            order.extend(list(self._vm.config.bootOptions.bootOrder))
            vm_spec.bootOptions = vim.vm.BootOptions(bootOrder=order)
        self._edit(vm_spec)  # Apply change to VM

    def _find_free_ide_controller(self):
        """
        Finds a free IDE controller to use
        :return: The free IDE controller
        :rtype: vim.vm.device.VirtualIDEController or None
        """
        for dev in self._vm.config.hardware.device:
            if isinstance(dev, vim.vm.device.VirtualIDEController) and len(dev.device) < 2:
                return dev  # If there are less than 2 devices attached, we can use it
        return None

    def mount_tools(self):
        """ Mounts the installer for VMware Tools """
        self._log.debug("Mounting tools installer on %s", self.name)
        self._vm.MountToolsInstaller().wait()

    def get_datastore_folder(self):
        """
        Gets the name of the VM's folder on the datastore
        :return: The name of the datastore folder with the VM's files
        :rtype: str
        """
        return os.path.split(self._vm.summary.config.vmPathName)[0]

    def get_hdd_by_name(self, name):
        """
        Gets a Virtual HDD from the VM
        :param name: Name of the virtual HDD
        :return: The HDD device
        :rtype: vim.vm.device.VirtualDisk or None
        """
        for dev in self._vm.config.hardware.device:
            if isinstance(dev, vim.vm.device.VirtualDisk) and \
                            dev.deviceInfo.label.lower() == name.lower():
                return dev
        return None

    def get_vim_vm(self):
        """
        Get the vim.VirtualMachine instance of the VM
        :return: The vim instance of the VM
        :rtype: vim.VirtualMachine
        """
        return self._vm

    def get_nics(self):
        """
        Returns a list of all Virtual Network Interface Cards (vNICs) on the VM
        :return: All vNICs on the VM
        :rtype: list(vim.vm.device.VirtualEthernetCard) or list
        """
        return [dev for dev in self._vm.config.hardware.device if is_vnic(dev)]

    def get_nic_by_name(self, name):
        """
        Gets a Virtual Network Interface Card from a VM
        :param str  name: Name of the vNIC
        :return: The vNIC found
        :rtype: vim.vm.device.VirtualEthernetCard or None
        """
        for dev in self._vm.config.hardware.device:
            if is_vnic(dev) and dev.deviceInfo.label.lower() == name.lower():
                return dev
        self._log.debug("Could not find vNIC '%s' on '%s'", name, self.name)
        return None

    def get_nic_by_id(self, nic_id):
        """
        Get a vNIC by integer ID
        :param int nic_id: ID of the vNIC
        :return: The vNIC found
        :rtype: vim.vm.device.VirtualEthernetCard or None
        """
        return self.get_nic_by_name("Network Adapter " + str(nic_id))

    def get_nic_by_network(self, network):
        """
        Finds a vNIC by it's network backing
        :param network: Network of the vNIC to match
        :type network: vim.Network
        :return: Name of the vNIC
        :rtype: str or None
        """
        for dev in self._vm.config.hardware.device:
            if is_vnic(dev) and dev.backing.network == network:
                return dev
        self._log.debug("Could not find vNIC with network '%s' on '%s'", network.name, self.name)
        return None

    def get_snapshot(self, snapshot=None):
        """
        Retrieves the named snapshot from the VM
        :param str snapshot: Name of the snapshot to get [default: current snapshot]
        :return: The snapshot found
        :rtype: vim.Snapshot or None
        """
        if snapshot is None:
            return self._vm.snapshot.currentSnapshot
        else:
            for snap in self.get_all_snapshots():
                if snap.name == snapshot:
                    return snap.snapshot
            return None

    def get_all_snapshots(self):
        """
        Retrieves a list of all snapshots of the VM
        :return: Nested List of vim.Snapshot objects
        :rtype: list(vim.Snapshot) or None
        """
        return self._get_snapshots_recursive(self._vm.snapshot.rootSnapshotList)

    # From: https://github.com/imsweb/ezmomi
    def _get_snapshots_recursive(self, snap_tree):
        """
        Recursively finds all snapshots in the tree
        :param snap_tree: Tree of snapshots
        :return: Nested List of vim.Snapshot objects
        :rtype: list(vim.Snapshot) or None
        """
        local_snap = []
        for snap in snap_tree:
            local_snap.append(snap)
        for snap in snap_tree:
            recurse_snap = self._get_snapshots_recursive(snap.childSnapshotList)
            if recurse_snap:
                local_snap.extend(recurse_snap)
        return local_snap

    def get_snapshot_info(self, name=None):
        """
        Human-readable info on a snapshot
        :param st name: Name of the snapshot to get [default: current snapshot]
        :return: Info on the snapshot found
        :rtype: str
        """
        snap = self.get_snapshot(name)
        return "\nName: %s; Description: %s; CreateTime: %s; State: %s" % \
               (snap.name, snap.description, snap.createTime, snap.state)

    def get_all_snapshots_info(self):
        """
        Enumerates the full snapshot tree of the VM and makes it human-readable
        :return: The human-readable snapshot tree info
        :rtype: str
        """
        # Check for ideas: snapshot_operations.py in pyvmomi_community_samples
        pass  # TODO: implement

    def get_info(self, detailed=False, uuids=False, snapshot=False, vnics=False):
        """
        Get human-readable information for a VM
        :param bool detailed: Add detailed information, such as maximum memory used [default: False]
        :param bool uuids: Whether to get UUID information [default: False]
        :param bool snapshot: Shows the current snapshot, if any [default: False]
        :param bool vnics: Add information about the virtual network interfaces on the VM
        :return: The VM's information
        :rtype: str
        """
        info_string = "\n"
        # http://pubs.vmware.com/vsphere-60/topic/com.vmware.wssdk.apiref.doc/vim.vm.Summary.html
        summary = self._vm.summary
        # http://pubs.vmware.com/vsphere-60/topic/com.vmware.wssdk.apiref.doc/vim.vm.ConfigInfo.html
        config = self._vm.config
        info_string += "Name          : %s\n" % self.name
        info_string += "Status        : %s\n" % str(summary.overallStatus)
        info_string += "Power State   : %s\n" % summary.runtime.powerState
        if self._vm.guest:
            info_string += "Guest State   : %s\n" % self._vm.guest.guestState
        info_string += "Last modified : %s\n" % str(self._vm.config.modified)  # datetime object
        if hasattr(summary.runtime, 'cleanPowerOff'):
            info_string += "Clean poweroff: %s\n" % summary.runtime.cleanPowerOff
        if detailed:
            info_string += "Num consoles  : %d\n" % summary.runtime.numMksConnections
        info_string += "Host          : %s\n" % self.host.name
        info_string += "Datastore     : %s\n" % self.datastore
        info_string += "HW Version    : %s\n" % config.version
        info_string += "Guest OS      : %s\n" % summary.config.guestFullName
        info_string += "Num CPUs      : %s\n" % summary.config.numCpu
        info_string += "Memory (MB)   : %s\n" % summary.config.memorySizeMB
        if detailed:
            info_string += "Num vNICs     : %s\n" % summary.config.numEthernetCards
            info_string += "Num Disks     : %s\n" % summary.config.numVirtualDisks
        info_string += "IsTemplate    : %s\n" % summary.config.template  # bool
        if detailed:
            info_string += "Config Path   : %s\n" % summary.config.vmPathName
        info_string += "Folder:       : %s\n" % self._vm.parent.name
        if self._vm.guest:
            info_string += "IP            : %s\n" % self._vm.guest.ipAddress
            info_string += "Hostname:     : %s\n" % self._vm.guest.hostName
            info_string += "Tools status  : %s\n" % self._vm.guest.toolsRunningStatus
            info_string += "Tools version : %s\n" % self._vm.guest.toolsVersionStatus2
        if vnics:
            vm_nics = self.get_nics()
            for num, vnic in zip(range(1, len(vm_nics) + 1), vm_nics):
                info_string += "vNIC %d label   : %s\n" % (num, vnic.deviceInfo.label)
                info_string += "vNIC %d summary : %s\n" % (num, vnic.deviceInfo.summary)
                info_string += "vNIC %d network : %s\n" % (num, vnic.backing.network.name)
        if uuids:
            info_string += "Instance UUID : %s\n" % summary.config.instanceUuid
            info_string += "Bios UUID     : %s\n" % summary.config.uuid
        if summary.runtime.question:
            info_string += "Question      : %s\n" % summary.runtime.question.text
        if summary.config.annotation:
            info_string += "Annotation    : %s\n" % summary.config.annotation
        if snapshot and self._vm.snapshot and hasattr(self._vm.snapshot, 'currentSnapshot'):
            info_string += "Current Snapshot: %s\n" % self._vm.snapshot.currentSnapshot.config.name
            info_string += "Disk usage of all snapshots: %s\n" % self.snapshot_disk_usage()
        if detailed and summary.runtime:
            info_string += "Last Poweron  : %s\n" % str(summary.runtime.bootTime)  # datetime object
            info_string += "Max CPU usage : %s\n" % summary.runtime.maxCpuUsage
            info_string += "Max Mem usage : %s\n" % summary.runtime.maxMemoryUsage
            info_string += "Last suspended: %s\n" % summary.runtime.suspendTime
        return info_string

    def screenshot(self):
        """
        Takes a screenshot of a VM
        :return: Path to datastore location of the screenshot
        :rtype: str
        """
        self._log.debug("Taking screenshot of %s", self.name)
        return self._vm.CreateScreenshot_Task().wait()

    def has_tools(self):
        """
        Checks if VMware Tools is installed and working
        :return: If tools are installed and working
        :rtype: bool
        """
        tools = self._vm.summary.guest.toolsStatus
        return True if tools == "toolsOK" or tools == "toolsOld" else False

    def powered_on(self):
        """
        Determines if a VM is powered on
        :return: If VM is powered on
        :rtype: bool
        """
        return self._vm.runtime.powerState == vim.VirtualMachine.PowerState.poweredOn

    def is_template(self):
        """
        Checks if VM is a template
        :return: If the VM is a template
        :rtype: bool
        """
        return bool(self._vm.summary.config.template)

    # http://pubs.vmware.com/vsphere-60/topic/com.vmware.wssdk.apiref.doc/vim.vm.GuestOsDescriptor.GuestOsIdentifier.html
    def is_windows(self):
        """
        Checks if a VM's guest OS is Windows
        :return: If guest OS is Windows
        :rtype: bool
        """
        return bool(str(self._vm.config.guestId).lower().startswith("win"))

    def _edit(self, config):
        """
        Reconfigures VM using the given configuration specification
        :param config: The configuration specification to apply
        :type config: vim.vm.ConfigSpec
        """
        if not self._vm.ReconfigVM_Task(config).wait():
            self._log.error("Failed to edit VM %s", self.name)

    def _customize(self, customization):
        """
        Customizes the VM using the given customization specification
        :param customization: The customization specification to apply
        :type customization: vim.vm.customization.Specification
        """
        if not self._vm.CheckCustomizationSpec(spec=customization).wait():
            self._log.error("Customization check failed for VM %s", self.name)
        elif not self._vm.CustomizeVM_Task(spec=customization).wait():
            self._log.error("Failed to customize VM %s", self.name)

    def __str__(self):
        return str(self.name)

    def __hash__(self):
        return hash(self._vm.summary.config.instanceUuid)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.name == other.name \
               and hash(self) == hash(other)

    def __ne__(self, other):
        return not self.__eq__(other)


def is_vnic(device):
    """
    Checks if the device is a VirtualEthernetCard
    :param device: The device to check
    :return: If the device is a vNIC
    :rtype: bool
    """
    return isinstance(device, vim.vm.device.VirtualEthernetCard)



# -*- coding: utf-8 -*-

from adles.vsphere import vsphere_utils
from adles.vsphere import folder_utils
from adles.vsphere import network_utils

from .vsphere_class import Vsphere
from .vm import VM
from .hosts import Host

__all__ = ['network_utils', 'vsphere_utils', 'folder_utils',
           'vsphere_class', 'vm', 'hosts', 'folder']



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from pyVmomi import vim


def create_portgroup(name, host, vswitch_name, vlan=0, promiscuous=False):
    """
    Creates a portgroup
    :param name: Name of portgroup to create
    :param host: vim.HostSystem on which to create the port group
    :param vswitch_name: Name of vSwitch on which to create the port group
    :param vlan: VLAN ID of the port group [default: 0]
    :param promiscuous: Put portgroup in promiscuous mode [default: False]
    """
    logging.debug("Creating PortGroup %s on vSwitch %s on host %s; VLAN: %d; Promiscuous: %s",
                  name, vswitch_name, host.name, vlan, promiscuous)
    policy = vim.host.NetworkPolicy()
    policy.security = vim.host.NetworkPolicy.SecurityPolicy()
    policy.security.allowPromiscuous = bool(promiscuous)
    policy.security.macChanges = False
    policy.security.forgedTransmits = False
    spec = vim.host.PortGroup.Specification(name=name, vlanId=int(vlan),
                                            vswitchName=vswitch_name, policy=policy)
    try:
        host.configManager.networkSystem.AddPortGroup(spec)
    except vim.fault.AlreadyExists:
        logging.error("PortGroup %s already exists on host %s", name, host.name)
    except vim.fault.NotFound:
        logging.error("vSwitch %s does not exist on host %s", vswitch_name, host.name)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from time import sleep, time

from pyVmomi import vim

SLEEP_INTERVAL = 0.05
LONG_SLEEP = 1.0


# TODO: could possibly use instead: from pyVim import WaitForTask
def wait_for_task(task, timeout=60.0, pause_timeout=True):
    """
    Waits for a single vCenter task to finish and returns its result
    :param task: The task to wait for
    :type task: vim.Task
    :param float timeout: Number of seconds to wait before terminating task [default: 60.0]
    :param bool pause_timeout: Pause timeout counter while task is queued on server [default: True]
    :return: Task result information (task.info.result)
    :rtype: str or None
    """
    if not task:  # Check if there's actually a task
        logging.error("No task was specified to wait for")
        return None
    name = str(task.info.descriptionId)
    obj = str(task.info.entityName)
    wait_time = 0.0
    end_time = time() + float(timeout)  # Set end time
    try:
        while True:
            if task.info.state == 'success':  # It succeeded!
                return task.info.result  # Return the task result if it was successful
            elif task.info.state == 'error':  # It failed...
                logging.error("Error during task %s on object '%s': %s",
                              name, obj, str(task.info.error.msg))
                return None
            elif time() > end_time:  # Check if it has exceeded the timeout
                logging.error("Task %s timed out after %s seconds", name, str(wait_time))
                task.CancelTask()  # Cancel the task since we've timed out
                return None
            elif task.info.state == 'queued':
                sleep(LONG_SLEEP)  # Sleep longer if it's queued up on system
                if pause_timeout is True:  # Don't count queue time against the timeout
                    end_time += LONG_SLEEP
                wait_time += LONG_SLEEP
            else:
                sleep(SLEEP_INTERVAL)  # Wait a bit so we don't waste resources checking state
                wait_time += SLEEP_INTERVAL
    except vim.fault.NoPermission as e:
        logging.error("Permission denied for task %s on %s: need privilege %s",
                      name, obj, e.privilegeId)
    except vim.fault.TaskInProgress as e:
        logging.error("Cannot complete task %s: task %s is already in progress on %s",
                      name, e.task.info.name, obj)
    except vim.fault.InvalidPowerState as e:
        logging.error("Cannot complete task %s: %s is in invalid power state %s",
                      name, obj, e.existingState)
    except vim.fault.InvalidState as e:
        logging.error("Cannot complete task %s: invalid state for %s\n%s", name, obj, str(e))
    except vim.fault.CustomizationFault:
        logging.error("Cannot complete task %s: invalid customization for %s", name, obj)
    except vim.fault.VmConfigFault:
        logging.error("Cannot complete task %s: invalid configuration for VM %s", name, obj)
    except vim.fault.InvalidName as e:
        logging.error("Cannot complete task %s for object %s: name '%s' is not valid",
                      name, obj, e.name)
    except vim.fault.DuplicateName as e:
        logging.error("Cannot complete task %s for %s: there is a duplicate named %s",
                      name, obj, e.name)
    except vim.fault.InvalidDatastore as e:
        logging.error("Cannot complete task %s for %s: invalid Datastore '%s'",
                      name, obj, e.datastore)
    except vim.fault.AlreadyExists:
        logging.error("Cannot complete task %s: %s already exists", name, obj)
    except vim.fault.NotFound:
        logging.error("Cannot complete task %s: %s does not exist", name, obj)
    except vim.fault.ResourceInUse:
        logging.error("Cannot complete task %s: resource %s is in use", name, obj)
    return None


# Possible alternate method(s) of doing tasks (TODO: profile performance):
#   https://github.com/vmware/pyvmomi-tools/blob/master/pyvmomi_tools/extensions/task.py
# This line allows calling "<task>.wait(<params>)" instead of "wait_for_task(task, params)"
# This works because the implicit first argument to a class method call in Python is the instance
vim.Task.wait = wait_for_task  # Inject into vim.Task class


# From: list_dc_datastore_info in pyvmomi-community-samples
def get_datastore_info(ds_obj):
    """
    Gets a human-readable summary of a Datastore
    :param ds_obj: The datastore to get information on
    :type ds_obj: vim.Datastore
    :return: The datastore's information
    :rtype: str
    """
    if not ds_obj:
        logging.error("No Datastore was given to get_datastore_info")
        return ""
    from adles.utils import sizeof_fmt
    info_string = "\n"
    summary = ds_obj.summary
    ds_capacity = summary.capacity
    ds_freespace = summary.freeSpace
    ds_uncommitted = summary.uncommitted if summary.uncommitted else 0
    ds_provisioned = ds_capacity - ds_freespace + ds_uncommitted
    ds_overp = ds_provisioned - ds_capacity
    ds_overp_pct = (ds_overp * 100) / ds_capacity if ds_capacity else 0

    info_string += "Name                  : %s\n" % summary.name
    info_string += "URL                   : %s\n" % summary.url
    info_string += "Capacity              : %s\n" % sizeof_fmt(ds_capacity)
    info_string += "Free Space            : %s\n" % sizeof_fmt(ds_freespace)
    info_string += "Uncommitted           : %s\n" % sizeof_fmt(ds_uncommitted)
    info_string += "Provisioned           : %s\n" % sizeof_fmt(ds_provisioned)
    if ds_overp > 0:
        info_string += "Over-provisioned      : %s / %s %%\n" \
                       % (sizeof_fmt(ds_overp), ds_overp_pct)
    info_string += "Hosts                 : %d\n" % len(ds_obj.host)
    info_string += "Virtual Machines      : %d" % len(ds_obj.vm)
    return info_string
vim.Datastore.get_info = get_datastore_info



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Perform Snapshot operations on Virtual Machines in a vSphere environment.

Usage:
    vm-snapshots [options]

Options:
    -h, --help          Prints this page
    --version           Prints current version
    -n, --no-color      Do not color terminal output
    -v, --verbose       Emit debugging logs to terminal
    -f, --file FILE     Name of JSON file with server connection information

Examples:
    vm-snapshots -vf logins.json

"""

import logging

from docopt import docopt

from adles.utils import script_setup, ask_question, resolve_path, is_vm
from adles.vsphere.folder_utils import format_structure
from adles.vsphere.vm import VM

__version__ = "0.2.0"


# noinspection PyUnboundLocalVariable
def main():
    args = docopt(__doc__, version=__version__, help=True)
    server = script_setup('cleanup_vms.log', args, (__file__, __version__))

    op = str(input("Enter Snapshot operation [create | revert | revert-current "
                   "| remove | remove-all | get | get-current | get-all | disk-usage]: "))
    if op == "create" or op == "revert" or op == "remove" or op == "get":
        name = str(input("Name of snapshot to %s: " % op))
        if op == "create":
            desc = str(input("Description of snapshot to create: "))
            memory = ask_question("Include memory?")
            quiesce = ask_question("Quiesce disks? (Requires VMware Tools to be on the VM)")
        elif op == "remove":
            children = ask_question("Remove any children of the snapshot?", default="yes")

    if ask_question("Multiple VMs? ", default="yes"):
        f, f_name = resolve_path(server, "folder", "with VMs")
        vms = [VM(vm=x) for x in f.childEntity if is_vm(x)]
        logging.info("Found %d VMs in folder '%s'", len(vms), f_name)
        if ask_question("Show the status of the VMs in the folder? "):
            logging.info("Folder structure: \n%s", format_structure(
                f.enumerate(recursive=True, power_status=True)))
        if not ask_question("Continue? ", default="yes"):
            logging.info("User cancelled operation, exiting...")
            exit(0)
    else:
        vms = [resolve_path(server, "vm", "to perform snapshot operations on")[0]]

    # Perform the operations
    for vm in vms:
        logging.info("Performing operation '%s' on VM '%s'", op, vm.name)
        if op == "create":
            vm.create_snapshot(name=name, description=desc, memory=memory, quiesce=quiesce)
        elif op == "revert":
            vm.revert_to_snapshot(snapshot=name)
        elif op == "revert-current":
            vm.revert_to_current_snapshot()
        elif op == "remove":
            vm.remove_snapshot(snapshot=name, remove_children=children)
        elif op == "remove-all":
            vm.remove_all_snapshots()
        elif op == "get":
            logging.info(vm.get_snapshot_info(name))    # TODO: implement
        elif op == "get-current":
            logging.info(vm.get_snapshot_info())        # TODO: implement
        elif op == "get-all":
            logging.info(vm.get_all_snapshots_info())   # TODO: implement
        elif op == "disk-usage":
            logging.info(vm.snapshot_disk_usage())
        else:
            logging.error("Unknown operation: %s", op)


if __name__ == '__main__':
    main()
